@@ -27,27 +27,28 @@ output_multi_t::output_multi_t(const std::string &name,
                           m_options.append, m_options.slim, m_options.droptemp,
                           m_options.hstore_mode, m_options.enable_hstore_index,
                           m_options.tblsmain_data, m_options.tblsmain_index)),
-      ways_pending_tracker(new id_tracker()), ways_done_tracker(new id_tracker()), rels_pending_tracker(new id_tracker()),
-      m_expire(new expire_tiles(&m_options)) {
-}
+      ways_done_tracker(new id_tracker()),
+      m_expire(m_options.expire_tiles_zoom, m_options.expire_tiles_max_bbox,
+               m_options.projection)
+{}
 
 output_multi_t::output_multi_t(const output_multi_t& other):
     output_t(other.m_mid, other.m_options), m_tagtransform(new tagtransform(&m_options)), m_export_list(new export_list(*other.m_export_list)),
     m_processor(other.m_processor), m_osm_type(other.m_osm_type), m_table(new table_t(*other.m_table)),
-    ways_pending_tracker(new id_tracker()), ways_done_tracker(new id_tracker()), rels_pending_tracker(new id_tracker()),
-    m_expire(new expire_tiles(&m_options)) {
-}
-
-
-output_multi_t::~output_multi_t() {
-}
-
-std::shared_ptr<output_t> output_multi_t::clone(const middle_query_t* cloned_middle) const{
-    output_multi_t *clone = new output_multi_t(*this);
-    clone->m_mid = cloned_middle;
     //NOTE: we need to know which ways were used by relations so each thread
     //must have a copy of the original marked done ways, its read only so its ok
-    clone->ways_done_tracker = ways_done_tracker;
+    ways_done_tracker(other.ways_done_tracker),
+    m_expire(m_options.expire_tiles_zoom, m_options.expire_tiles_max_bbox,
+             m_options.projection)
+{}
+
+
+output_multi_t::~output_multi_t() = default;
+
+std::shared_ptr<output_t> output_multi_t::clone(const middle_query_t* cloned_middle) const
+{
+    auto *clone = new output_multi_t(*this);
+    clone->m_mid = cloned_middle;
     return std::shared_ptr<output_t>(clone);
 }
 
@@ -57,11 +58,11 @@ int output_multi_t::start() {
 }
 
 size_t output_multi_t::pending_count() const {
-    return ways_pending_tracker->size() + rels_pending_tracker->size();
+    return ways_pending_tracker.size() + rels_pending_tracker.size();
 }
 
 void output_multi_t::enqueue_ways(pending_queue_t &job_queue, osmid_t id, size_t output_id, size_t& added) {
-    osmid_t const prev = ways_pending_tracker->last_returned();
+    osmid_t const prev = ways_pending_tracker.last_returned();
     if (id_tracker::is_valid(prev) && prev >= id) {
         if (prev > id) {
             job_queue.push(pending_job_t(id, output_id));
@@ -77,7 +78,7 @@ void output_multi_t::enqueue_ways(pending_queue_t &job_queue, osmid_t id, size_t
     }
 
     //grab the first one or bail if its not valid
-    osmid_t popped = ways_pending_tracker->pop_mark();
+    osmid_t popped = ways_pending_tracker.pop_mark();
     if(!id_tracker::is_valid(popped))
         return;
 
@@ -87,11 +88,11 @@ void output_multi_t::enqueue_ways(pending_queue_t &job_queue, osmid_t id, size_t
             job_queue.push(pending_job_t(popped, output_id));
             added++;
         }
-        popped = ways_pending_tracker->pop_mark();
+        popped = ways_pending_tracker.pop_mark();
     }
 
     //make sure to get this one as well and move to the next
-    if(popped == id) {
+    if (popped > id) {
         if (!ways_done_tracker->is_marked(popped) && id_tracker::is_valid(popped)) {
             job_queue.push(pending_job_t(popped, output_id));
             added++;
@@ -114,7 +115,7 @@ int output_multi_t::pending_way(osmid_t id, int exists) {
 }
 
 void output_multi_t::enqueue_relations(pending_queue_t &job_queue, osmid_t id, size_t output_id, size_t& added) {
-    osmid_t const prev = rels_pending_tracker->last_returned();
+    osmid_t const prev = rels_pending_tracker.last_returned();
     if (id_tracker::is_valid(prev) && prev >= id) {
         if (prev > id) {
             job_queue.push(pending_job_t(id, output_id));
@@ -130,7 +131,7 @@ void output_multi_t::enqueue_relations(pending_queue_t &job_queue, osmid_t id, s
     }
 
     //grab the first one or bail if its not valid
-    osmid_t popped = rels_pending_tracker->pop_mark();
+    osmid_t popped = rels_pending_tracker.pop_mark();
     if(!id_tracker::is_valid(popped))
         return;
 
@@ -138,11 +139,11 @@ void output_multi_t::enqueue_relations(pending_queue_t &job_queue, osmid_t id, s
     while (popped < id) {
         job_queue.push(pending_job_t(popped, output_id));
         added++;
-        popped = rels_pending_tracker->pop_mark();
+        popped = rels_pending_tracker.pop_mark();
     }
 
     //make sure to get this one as well and move to the next
-    if(popped == id) {
+    if (popped > id) {
         if(id_tracker::is_valid(popped)) {
             job_queue.push(pending_job_t(popped, output_id));
             added++;
@@ -163,10 +164,13 @@ int output_multi_t::pending_relation(osmid_t id, int exists) {
     return ret;
 }
 
-void output_multi_t::stop() {
+void output_multi_t::stop()
+{
     m_table->stop();
-    m_expire->output_and_destroy();
-    m_expire.reset();
+    if (m_options.expire_tiles_zoom_min >= 0) {
+        m_expire.output_and_destroy(m_options.expire_tiles_filename.c_str(),
+                                    m_options.expire_tiles_zoom_min);
+    }
 }
 
 void output_multi_t::commit() {
@@ -269,7 +273,7 @@ int output_multi_t::process_node(osmid_t id, double lat, double lon, const tagli
         //grab its geom
         auto geom = m_processor->process_node(lat, lon);
         if (geom.valid()) {
-            m_expire->from_bbox(lon, lat, lon, lat);
+            m_expire.from_bbox(lon, lat, lon, lat);
             copy_node_to_table(id, geom.geom, outtags);
         }
     }
@@ -284,7 +288,7 @@ int output_multi_t::reprocess_way(osmid_t id, const nodelist_t &nodes, const tag
         way_delete(id);
         const std::vector<osmid_t> rel_ids = m_mid->relations_using_way(id);
         for (std::vector<osmid_t>::const_iterator itr = rel_ids.begin(); itr != rel_ids.end(); ++itr) {
-            rels_pending_tracker->mark(*itr);
+            rels_pending_tracker.mark(*itr);
         }
     }
 
@@ -320,7 +324,7 @@ int output_multi_t::process_way(osmid_t id, const idlist_t &nodes, const taglist
             //if we are also interested in relations we need to mark
             //this way pending just in case it shows up in one
             if (m_processor->interests(geometry_processor::interest_relation)) {
-                ways_pending_tracker->mark(id);
+                ways_pending_tracker.mark(id);
             } else {
                 // We wouldn't be interested in this as a relation, so no need to mark it pending.
                 // TODO: Does this imply anything for non-multipolygon relations?
@@ -381,7 +385,7 @@ int output_multi_t::process_relation(osmid_t id, const memberlist_t &members,
             for (const auto geom: geoms) {
                 //TODO: we actually have the nodes in the m_relation_helper and could use them
                 //instead of having to reparse the wkb in the expiry code
-                m_expire->from_wkb(geom.geom.c_str(), -id);
+                m_expire.from_wkb(geom.geom.c_str(), -id);
                 //what part of the code relies on relation members getting negative ids?
                 copy_to_table(-id, geom, outtags, make_polygon);
             }
@@ -422,7 +426,7 @@ void output_multi_t::copy_to_table(const osmid_t id, const geometry_builder::pg_
     if (geom.is_polygon()) {
         // It's a polygon table (implied by it turning into a poly),
         // and it got formed into a polygon, so expire as a polygon and write the geom
-        m_expire->from_nodes_poly(m_way_helper.node_cache, id);
+        m_expire.from_nodes_poly(m_way_helper.node_cache, id);
         if (geom.area > 0.0) {
             char tmp[32];
             snprintf(tmp, sizeof(tmp), "%g", geom.area);
@@ -433,33 +437,34 @@ void output_multi_t::copy_to_table(const osmid_t id, const geometry_builder::pg_
         // Linestring
         if (!polygon) {
             // non-polygons are okay
-            m_expire->from_nodes_line(m_way_helper.node_cache);
+            m_expire.from_nodes_line(m_way_helper.node_cache);
             m_table->write_row(id, tags, geom.geom);
         }
     }
 }
 
 void output_multi_t::delete_from_output(osmid_t id) {
-    if(m_expire->from_db(m_table.get(), id))
+    if(m_expire.from_db(m_table.get(), id))
         m_table->delete_row(id);
 }
 
-void output_multi_t::merge_pending_relations(std::shared_ptr<output_t> other) {
-    std::shared_ptr<id_tracker> tracker = other.get()->get_pending_relations();
-    osmid_t id;
-    while(tracker.get() && id_tracker::is_valid((id = tracker->pop_mark()))){
-        rels_pending_tracker->mark(id);
+void output_multi_t::merge_pending_relations(output_t *other)
+{
+    auto *omulti = dynamic_cast<output_multi_t *>(other);
+
+    if (omulti) {
+        osmid_t id;
+        while (id_tracker::is_valid((id = omulti->rels_pending_tracker.pop_mark()))) {
+            rels_pending_tracker.mark(id);
+        }
     }
 }
 
-void output_multi_t::merge_expire_trees(std::shared_ptr<output_t> other) {
-    if(other->get_expire_tree().get())
-        m_expire->merge_and_destroy(*other.get()->get_expire_tree());
-}
+void output_multi_t::merge_expire_trees(output_t *other)
+{
+    auto *omulti = dynamic_cast<output_multi_t *>(other);
 
-std::shared_ptr<id_tracker> output_multi_t::get_pending_relations() {
-    return rels_pending_tracker;
-}
-std::shared_ptr<expire_tiles> output_multi_t::get_expire_tree() {
-    return m_expire;
+    if (omulti) {
+        m_expire.merge_and_destroy(omulti->m_expire);
+    }
 }
